import re
import subprocess


def to_cli_args(*args, **kwargs):
    cmd = []
    for k, v in kwargs.items():
        k = k.replace('_', '-')
        short = len(k) == 1
        if short:
            cmd.append('-' + k)
            if v is not True:
                cmd.append(v)
        else:
            if v is True:
                cmd.append('--' + k)
            else:
                cmd.append('--{0}={1}'.format(k, v))

    cmd.extend(args)
    return cmd


class Git(object):
    def __init__(self, cmd=None, repo_path=None):
        self.repo_path = repo_path
        self.current = list(cmd) if cmd else ['git']

    def __call__(self, *args, **kwargs):
        cmd = self.current + to_cli_args(*args, **kwargs)
        subprocess_kwargs = {}
        if self.repo_path:
            subprocess_kwargs['cwd'] = self.repo_path
        return subprocess.check_output(cmd, **subprocess_kwargs).decode()

    def __getattr__(self, name):
        name = name.replace('_', '-')
        return Git(self.current + [name], repo_path=self.repo_path)

    def __str__(self):
        return str(self.current)

    def __repr__(self):
        return '<Git {}>'.format(str(self))


class Status(object):
    def __init__(self, state=None, is_modified=False, path=None, original_path=None):
        self.state = state
        self.path = path
        self.original_path = original_path
        self.is_modified = is_modified

    def __str__(self):
        status = '{:1}{:1}'.format(self.state, 'M' if self.is_modified else '')
        if self.is_renamed:
            status = '{:3}{} -> {}'.format(status, self.original_path, self.path)
        else:
            status = '{:3}{}'.format(status, self.path)
        return status

    def __repr__(self):
        return '<Status {}>'.format(self.path)

    @classmethod
    def from_string(cls, string):
        string = string  # Ensure that the string is actually a string
        status = cls()

        status.state = string[0]
        status.is_modified = string[1] == 'M'

        if status.is_renamed:
            matches = re.search('(\S+?)\s+->\s+(\S+?)$', string[3:])
            status.original_path = matches.groups()[0]
            status.path = matches.groups()[1]
        else:
            status.path = string[3:]

        return status

    @property
    def is_staged(self):
        """If the state is empty then the file is unstaged."""
        return self.state not in (' ', '?',)

    @property
    def is_untracked(self):
        """Returns true if the file is untracked."""
        return self.state == '?'

    @property
    def is_renamed(self):
        """Returns true if the file is renamed."""
        return self.state == 'R'

    @property
    def is_deleted(self):
        """Returns true if the file is deleted."""
        return self.state == 'D'

    @property
    def is_added(self):
        """Returns true if the file is added."""
        return self.state == 'A'



#!/usr/bin/env python
import os

from therapist import Runner
from therapist.printer import Printer


BASE_DIR = os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
printer = Printer()


try:
    runner = Runner(os.path.join(BASE_DIR, '.therapist.yml'))
except Runner.Misconfigured as err:
    printer.fprint('Misconfigured: '.format(err.message), 'red')
    exit(1)
except Runner.UnstagedChanges as err:
    printer.fprint(err.message, 'red')
    exit(1)

try:
    runner.run()
except runner.ActionFailed:
    exit(1)



from six import print_


class Printer(object):
    STYLES = {
        'RESET': '\x1b[0m',
        'BOLD': '\x1b[1m',
        'UNDERLINE': '\x1b[4m',
        'BLINK': '\x1b[5m',
        'BLACK': '\x1b[30m',
        'RED': '\x1b[31m',
        'GREEN': '\x1b[32m',
        'YELLOW': '\x1b[33m',
        'BLUE': '\x1b[34m',
        'MAGENTA': '\x1b[35m',
        'CYAN': '\x1b[36m',
        'GRAY': '\x1b[37m',
    }

    def format(self, text, *styles):
        """Returns a string with the styles applied."""
        for style in styles:
            if style.upper() in self.STYLES:
                text = '%s%s' % (self.STYLES[style.upper()], text)

        text = '%s%s%s' % (self.STYLES['RESET'], text, self.STYLES['RESET'])

        return text

    def fprint(self, text='', *styles, **kwargs):
        """Prints a string with the styles applied."""
        end = '' if kwargs.get('inline', False) else '\n'
        print_(self.format(text, *styles), end=end)



import os
import stat

from shutil import copy

import click

from therapist import Runner
from therapist._version import __version__
from therapist.printer import Printer


MODE_775 = (stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR |
            stat.S_IRGRP | stat.S_IWGRP | stat.S_IXGRP |
            stat.S_IROTH | stat.S_IXOTH)

printer = Printer()


def find_git_dir():
    """Locate the .git directory."""
    path = os.path.abspath(os.curdir)
    while path != os.path.abspath('/'):
        if os.path.isdir(os.path.join(path, '.git')):
            return os.path.join(path, '.git')
        path = os.path.dirname(path)
    return None


@click.group()
@click.version_option(__version__, prog_name='therapist')
def cli():
    """A smart pre-commit hook for git."""
    pass


@cli.command()
def install():
    """Install the pre-commit hook."""
    srchook = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'hooks', 'pre-commit.py')
    gitdir_path = find_git_dir()

    if gitdir_path is None:
        printer.fprint('Unable to locate git repo.', 'red')
        exit(1)

    dsthook = os.path.join(gitdir_path, 'hooks', 'pre-commit')

    if os.path.isfile(dsthook):
        printer.fprint('There is an existing pre-commit hook.', 'yellow')

        if not click.confirm('Are you sure you want to overwrite this hook?'):
            printer.fprint('Installation aborted.')
            exit(1)

    printer.fprint('Installing pre-commit hook...', inline=True)
    copy(srchook, dsthook)
    os.chmod(dsthook, MODE_775)
    printer.fprint('\tDONE', 'green', 'bold')


@cli.command()
@click.argument('paths', nargs=-1)
@click.option('--action', '-a', default=None, help='A name of a specific action to be run.')
@click.option('--include-unstaged', is_flag=True, help='Include unstaged files.')
@click.option('--include-untracked', is_flag=True, help='Include untracked files.')
def run(paths, action, include_unstaged, include_untracked):
    """Run actions as a batch or individually."""
    gitdir_path = find_git_dir()

    if gitdir_path is None:
        printer.fprint('Unable to locate git repo.', 'red')
        exit(1)

    repo_root = os.path.dirname(gitdir_path)

    # If paths were provided get all the files for each path
    if paths:
        files = []
        for path in paths:
            if os.path.isdir(path):
                for stats in os.walk(path):
                    for f in stats[2]:
                        files.append(os.path.abspath(os.path.join(stats[0], f)))
            else:
                files.append(os.path.abspath(path))
    else:
        files = None

    try:
        runner = Runner(os.path.join(repo_root, '.therapist.yml'), files=files, ignore_unstaged_changes=True,
                        include_unstaged=include_unstaged, include_untracked=include_untracked)
    except Runner.Misconfigured as err:
        printer.fprint('Misconfigured: '.format(err.message), 'red')
        exit(1)

    try:
        if action:
            printer.fprint()
            try:
                runner.run_action(action)
            except runner.ActionDoesNotExist as err:
                printer.fprint(err.message)
                printer.fprint()
                printer.fprint('Available actions:')

                for a in runner.actions:
                    printer.fprint(a)
            printer.fprint()
        else:
            runner.run()
    except runner.ActionFailed:
        exit(1)


@cli.command()
def uninstall():
    """Uninstall the current pre-commit hook."""
    gitdirpath = find_git_dir()

    if gitdirpath is None:
        printer.fprint('Unable to locate git repo.', 'red')
        exit(1)

    hookpath = os.path.join(gitdirpath, 'hooks', 'pre-commit')

    if not os.path.isfile(hookpath):
        printer.fprint('There is no hook to uninstall.', 'yellow')
        exit(1)

    if click.confirm('Are you sure you want to uninstall the current pre-commit hook?'):
        printer.fprint('Uninstalling pre-commit hook...', inline=True)
        os.remove(hookpath)
        printer.fprint('\tDONE', 'green', 'bold')



import os
import subprocess
import yaml

from distutils.spawn import find_executable
from pathspec import GitIgnorePattern, PathSpec

from therapist.git import Git, Status
from therapist.printer import Printer

printer = Printer()

GIT_BINARY = find_executable('git')


def execute_action(action, files, cwd):
    if 'include' in action:
        rules = action['include'] if isinstance(action['include'], list) else [action['include']]
        spec = PathSpec(map(GitIgnorePattern, rules))
        files = list(spec.match_files(files))

    if 'exclude' in action:
        rules = action['exclude'] if isinstance(action['exclude'], list) else [action['exclude']]
        spec = PathSpec(map(GitIgnorePattern, rules))
        exclude = list(spec.match_files(files))
        files = list(filter(lambda f: f not in exclude, files))

    if 'run' in action and files:
        file_list = ' '.join(files)
        run_command = action['run'].format(files=file_list)

        try:
            pipes = subprocess.Popen(run_command, shell=True, cwd=cwd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        except OSError as err:
            return 'OSError {}'.format(str(err)), Runner.ERROR

        std_out, std_err = pipes.communicate()

        if len(std_err):
            return std_err, Runner.FAILURE
        else:
            return std_out, Runner.SUCCESS if pipes.returncode == 0 else Runner.FAILURE

    return None, Runner.SKIPPED


class Runner(object):
    SUCCESS = 0
    FAILURE = 1
    SKIPPED = 2
    ERROR = 3

    class ActionDoesNotExist(Exception):
        def __init__(self, message='', *args, **kwargs):
            self.message = message
            super(self.__class__, self).__init__(*args, **kwargs)

    class Misconfigured(Exception):
        NO_CONFIG_FILE = 1
        NO_ACTIONS = 2

        def __init__(self, message='', code=0, *args, **kwargs):
            self.message = message
            self.code = code
            super(self.__class__, self).__init__(*args, **kwargs)

    class UnstagedChanges(Exception):
        def __init__(self, message='', *args, **kwargs):
            self.message = message
            super(self.__class__, self).__init__(*args, **kwargs)

    class ActionFailed(Exception):
        def __init__(self, *args, **kwargs):
            self.actions = args
            super(self.__class__, self).__init__(*args, **kwargs)

    def __init__(self, config_path, files=None, ignore_unstaged_changes=False, include_unstaged=False,
                 include_untracked=False):
        self.cwd = os.path.abspath(os.path.dirname(config_path))
        self.git = Git(repo_path=self.cwd)

        # Try and load the config file
        try:
            with open(config_path, 'r') as f:
                config = yaml.safe_load(f)
                f.close()
        except IOError:
            raise self.Misconfigured('Missing configuration file.', code=self.Misconfigured.NO_CONFIG_FILE)
        else:
            if 'actions' in config:
                self.actions = config['actions']
            else:
                raise self.Misconfigured('`actions` was not specified in the configuration file.',
                                         code=self.Misconfigured.NO_ACTIONS)

        if files:
            self.files = files
        else:
            self.files = []

            args = []

            if not include_untracked:
                args.append('-uno')

            status = self.git.status(*args, porcelain=True)

            for line in status.splitlines():
                file_status = Status.from_string(line)

                # Check if staged files were modified since being staged
                if file_status.is_staged and file_status.is_modified and not ignore_unstaged_changes:
                    raise self.UnstagedChanges('There are unstaged changes.')

                # Skip unstaged files if the `unstaged` flag is False
                if not file_status.is_staged and not include_unstaged and not include_untracked:
                    continue

                # Skip deleted files
                if file_status.is_deleted:
                    continue

                self.files.append(file_status.path)

    def run_action(self, action_name):
        """Runs a single action."""
        failures = []
        errors = []

        action = self.actions.get(action_name, None)

        if not action:
            raise self.ActionDoesNotExist('`{}` is not a valid action.'.format(action_name))

        description = '%s ' % action.get('description', action_name)[:68]
        printer.fprint(description.ljust(69, '.'), 'bold', inline=True)

        output, status = execute_action(action, self.files, self.cwd)

        if status == self.SUCCESS:
            printer.fprint(' [SUCCESS]', 'green', 'bold')
        elif status == self.FAILURE:
            printer.fprint(' [FAILURE]', 'red', 'bold')
            failures.append({'description': description, 'output': output})
        elif status == self.ERROR:
            printer.fprint('..', 'bold', inline=True)
            printer.fprint(' [ERROR]', 'red', 'bold')
            errors.append({'description': description, 'output': output})
        else:
            printer.fprint(' [SKIPPED]', 'cyan', 'bold')

        # Helper function to print reports
        def _print_report(title, report):
            if report:
                printer.fprint()
                printer.fprint(''.ljust(79, '='), 'bold')
                printer.fprint(title[:79], 'bold')
                printer.fprint(''.ljust(79, '='), 'bold')
                printer.fprint(report)

        for failure in failures:
            _print_report('FAILED: ' + failure['description'], failure['output'])

        for error in errors:
            _print_report('ERROR: ' + error['description'], error['output'])

        if failures or errors:
            raise self.ActionFailed(action_name)

    def run(self):
        """Runs the set of actions."""
        if self.files:
            failures = []

            printer.fprint()

            # Run each actions
            for name in self.actions:
                try:
                    self.run_action(name)
                except self.ActionFailed:
                    failures.append(name)

            printer.fprint()

            if failures:
                raise self.ActionFailed(*failures)



"""
Smart pre-commit hook for git.
"""
import os
import re

from setuptools import find_packages, setup


DEPENDENCIES = ['click', 'pathspec', 'PyYAML', 'six']
ROOT = os.path.abspath(os.path.dirname(__file__))


def get_version():
    versionfile = os.path.join(ROOT, 'therapist', '_version.py')
    verstrline = open(versionfile, 'rt').read()
    mo = re.search(r"""^__version__ = ['"]([^'"]*)['"]""", verstrline, re.M)
    if mo:
        return mo.group(1)
    else:
        raise RuntimeError('Unable to find version string in {0}.'.format(versionfile))

setup(
    name='therapist',
    version=get_version(),
    url='https://github.com/rehandalal/therapist',
    license='Mozilla Public License Version 2.0',
    author='Rehan Dalal',
    author_email='rehan@meet-rehan.com',
    description='A smart pre-commit hook for git.',
    long_description=__doc__,
    packages=find_packages(exclude=['tests', 'docs']),
    include_package_data=True,
    zip_safe=False,
    platforms='any',
    install_requires=DEPENDENCIES,
    setup_requires=['pytest-runner'],
    tests_require=['pytest'],
    py_modules=['therapist'],
    entry_points={
        'console_scripts': [
            'therapist = therapist.cli:cli',
        ],
    },
    classifiers=[
        # As from http://pypi.python.org/pypi?%3Aaction=list_classifiers
        # 'Development Status :: 1 - Planning',
        # 'Development Status :: 2 - Pre-Alpha',
        # 'Development Status :: 3 - Alpha',
        'Development Status :: 4 - Beta',
        # 'Development Status :: 5 - Production/Stable',
        # 'Development Status :: 6 - Mature',
        # 'Development Status :: 7 - Inactive',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: MacOS',
        'Operating System :: Unix',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ]
)
